/* Copyright (c) 2007-2008 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/**
 * Source Code Control javascript API.
 *
 * TODO:
 *   * All of the SCC functions in ko.scc require a koISCC service to use.
 *     It would be nice if this service was not required and then determine
 *     the appropriate service from the URL itself.
 */

if (typeof(ko)=='undefined') {
    var ko = {};
}

/* Generic SCC API */
ko.scc = {};
(function() {

    /*-----------------   Internal   -----------------*/

        /**
         * Internal logging control used for this namespace.
         */
    var log = ko.logging.getLogger('SCC');

        /**
         * An array of all scc XPCOM services. Each array item contains a
         * dictionary of: { "name": component_name, "component": scc_service }
         */
    var _sccAllComponents = null;

        /**
         * Komodo prefs.
         * @type Components.interfaces.koIOrderedPreference
         */
    var _prefs = Components.classes["@activestate.com/koPrefService;1"].
                        getService(Components.interfaces.koIPrefService).prefs;

        /**
         * Internal helper function to retrieve the SCC diff options.
         * @private
         */
    function _getDiffOptions(sccSvc) {
        var options = '';

        // p4 has a special setup for performing diff operations, all others
        // simply use the diff options preference.
        if (sccSvc.name != 'p4' ||
            _prefs.getStringPref('p4_diff_type') == 'komododiff') {
            if (_prefs.hasStringPref(sccSvc.name+'DiffOptions')) {
                options = _prefs.getStringPref(sccSvc.name+'DiffOptions');
            }
        }
        return options;
    };

        /**
         * Internal helper function to run an asynchronous scc command.
         * @private
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param cmdName {string}  The name of the scc command being run.
         * @param urls {array}  The list of urls used.
         * @param postCallback {function}  Function to be called on completion.
         */
    function _genericAsyncSccCommand(sccSvc, cmdName, urls, postCallback) {
        var _log = log;
        var async_callback = {
            "callback": function(result, data) {
                try {
                    // data will be an array of koISCCHistoryItem,
                    // throw it into the tree view.
                    if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {

                        ko.scc.logger.reportSCCAction(sccSvc, cmdName, urls, data);

                        if (cmdName == "revert") {
                            ko.views.manager.revertViewsByURL(urls);
                            // handle reverting projects properly
                            for (var i in urls) {
                                if (urls[i].slice(-4).toLowerCase() == '.kpf') {
                                    ko.projects.manager.revertProjectByURL(urls[i]);
                                }
                            }
                        }

                    } else if (result == Components.interfaces.koIAsyncCallback.RESULT_ERROR) {
                        // data should contain the error message then.
                        var title = sccSvc.name + " " + cmdName + " failure";
                        if (!data) {
                            data = cmdName + " could not be completed.";
                        }
                        ko.dialogs.alert(title, data, title);
                    }

                } catch (e) {
                    _log.warn("update callback failed, exception: " + e);
                } finally {
                    _log = null;
                    // If supplied, trigger the callback function.
                    if (postCallback) {
                        postCallback(result, data);
                    }
                }
            }
        }

        if (cmdName == "edit") {
            sccSvc.edit(urls.length, urls, async_callback);
        } else if (cmdName == "add") {
            sccSvc.add(urls.length, urls, '', '', async_callback);
        } else if (cmdName == "remove") {
            var force = true;
            var recursive = false;
            sccSvc.remove(urls.length, urls, force, recursive, async_callback);
        } else if (cmdName == "revert") {
            sccSvc.revert(urls.length, urls, '', async_callback);
        }
    };

        /**
         * Handle output from SCC update and report the results.
         * @param changes {string}  The output from the scc update command.
         */
    function _handleChanges(changes) {
        //XXX This pattern is specific to CVS update output. What about Perforce
        //    sync output? Is this a bug or is .handleChanges only called for CVS
        //    updates? --TrentM

        // re ^(?P<status>[UPARMC\?]) (?P<file>.*)$
        RegExp.multiline = true;
        var pattern = new RegExp("^([UCDP\?])\s+(.*?)\s+(.*?)$","g");
        var result;
        var conflicts = [];
        var haveConflict = false;
        var removed = [];
        var haveRemoved = false;
        var added = [];
        var haveNew = false;
        var file;
        var list;
        var i,j;
        while ((result = pattern.exec(changes)) != null) {
            //dump('found: '+result[1]+' '+result[2]+' '+result[3]+'\n');
            switch (result[1]) {
            case 'C':
                // conflict
                haveConflict = true;
                conflicts.push(result[3]);
                break;
            case 'D':
                // removed
                haveRemoved = true;
                removed.push(result[3]);
                break;
            case 'P':
                break; // Patched file can't be a new file, even though it may not be in our project.
            case 'U':
                // either updated or new file --
                // XXX need to check to make sure that it really is new -- we add
                // all files that aren't in the project
                if (!ko.projects.hasURL(result[3])) {
                    haveNew = true;
                    added.push(result[3]);
                }
                break;
            }
        }
        // look to see if files were added, ask if they should be
        // added to the current project.
        // XXX FIXME this needs to add for each project, not just the current
        var filepart, selectedURLs, obj;
        if (haveNew && ko.projects.manager.getCurrentProject()) {
            //XXX I think either (1) we should 'allowNone' here so the user could
            //    chose to not add new files, yet continue on to remove files, or
            //    better yet (2) we should group the added and removed files and
            //    ask the user to select which updates to carry to their project in
            //    one dialog.
            selectedURLs = ko.dialogs.selectFromList(
                    "Add New Files", // title
                    "Some new files from the repository are not in your "+
                        "project.  Select the files to add to your project.",
                    added,
                    null, // selectionCondition (null means default)
                    ko.uriparse.displayPath); // stringifier
            if (selectedURLs == null) { // user cancelled dialog
                return false;
            }

            for (i = 0; i < selectedURLs.length; i++) {
                filepart = ko.projects.manager.getCurrentProject().createPartFromType('file');
                filepart.setStringAttribute('url', selectedURLs[i]);
                ko.projects.manager.addItem(filepart);
            }
        }
        // look to see if files were removed, notify of removal,
        // ask to remove from projects.
        if (haveRemoved) {
            var aroundToRemove = [];
            for (i = 0; i < removed.length; i++) {
                list = ko.projects.findPartsByURL(removed[i]);
                if (list.length > 0 && !list[0].live) {
                    aroundToRemove.push(selectedURLs[i]);
                }
            }

            selectedURLs = ko.dialogs.selectFromList(
                    "Remove Old Files", // title
                    "Some files have been removed from the repository. "+
                        "Select the files to remove from your projects and "+
                        "toolbox.",
                    aroundToRemove,
                    null, // selectionCondition (null means default)
                    ko.uriparse.displayPath); // stringifier
            if (selectedURLs == null) { // user cancelled dialog
                return false;
            }
            ko.projects.removeItemsByURLList(selectedURLs);
        }

        // look to see if there are conflicts, notify of conflicts,
        // and ask to open the files in komodo.  also mark any
        // items with conflict status
        if (haveConflict) {
            selectedURLs = ko.dialogs.selectFromList(
                    "Merge Conflicts", // title
                    "There are conflicts in some updated file(s). "+
                        "Would you like to open those file(s)?",
                    conflicts,
                    null, // selectionCondition (null means default)
                    ko.uriparse.displayPath); // stringifier
            if (selectedURLs == null) { // user cancelled dialog
                return false;
            }
            ko.open.multipleURIs(selectedURLs);
        }

        // XXX reset the static var, is this necessary?
        RegExp.multiline = false;
        return true;
    };


    /*-----------------   External   -----------------*/

        /**
         * Get the available scc names and components.
         *
         * @param onlyEnabled {bool}  Whether to only return functional
         *                            components (the default is false)
         * @returns {Array}  list of objects containing "name" and "component"
         */
    this.getAvailableSCCComponents = function KoSccGetAvailableSCCComponents(onlyEnabled /*false */) {
        if (_sccAllComponents == null) {
            _sccAllComponents = [];
            var catman = Components.classes["@mozilla.org/categorymanager;1"].
                            getService(Components.interfaces.nsICategoryManager);
            var category = 'category-komodo-scc';
            var names = catman.enumerateCategory(category);
            var nameObj;
            var name;
            var cid;
            var scc_service;
            while (names.hasMoreElements()) {
                nameObj = names.getNext();
                nameObj.QueryInterface(Components.interfaces.nsISupportsCString);
                name = nameObj.data;
                //cid = catman.getCategoryEntry(category, name);
                cid = "@activestate.com/koSCC?type=" + name + ";1";
                try {
                    scc_service = Components.classes[cid].
                                    getService(Components.interfaces.koISCC);
                    _sccAllComponents.push({"name": name, "component": scc_service});
                } catch(ex) {
                    log.exception(ex, "Unable to get " + name + " scc component");
                }
            }
        }
        var scc_components = _sccAllComponents;
        var sccSvc;
        if (onlyEnabled) {
            scc_components = [];
            for (var i=0; i < _sccAllComponents.length; i++) {
                sccSvc = _sccAllComponents[i].component;
                if (!sccSvc.isFunctional || !sccSvc.isEnabled) {
                    continue;
                }
                scc_components.push(sccSvc);
            }
        }
        return scc_components;
    };


        /**
         * Return the SCC services that can handles the supplied koIFile.
         * If Komodo has not yet checked the supplied file for SCC support
         * then it is possible this function will return with no matching
         * service even though the file may be under a SCC system.
         * XXX - How to avoid this, force the fileStatusService to check
         *       the file?
         * @param koIFile {Components.interfaces.koIFileEx}  File needing SCC support.
         * @returns {Components.interfaces.koISCC}  SCC service instance.
         */
    this.getServiceForFile = function KoSccGetServiceForFile(koIFile) {
        try {
            var sccType = koIFile.sccType || koIFile.sccDirType;
            if (!sccType || koIFile.sccExclude) {
                return null;
            }

            var cid = "@activestate.com/koSCC?type=" + sccType + ";1";
            return Components.classes[cid].getService(Components.interfaces.koISCC);
        } catch (e) {
            log.exception(e);
        }
        return null;
    };

        /**
         * Return the SCC services that can handles the supplied koIFile.
         * If Komodo has not yet checked the supplied url/file for SCC support
         * then it is possible this function will return with no matching
         * service even though the file may be under a SCC system.
         * XXX - How to avoid this, force the fileStatusService to check
         *       the file?
         * @param url {wstring}  Url to get the handling SCC services for.
         * @returns {Components.interfaces.koISCC}  SCC service instance.
         */
    this.getServiceForUrl = function KoSccGetServiceForUrl(url) {
        try {
            var fileService = Components.classes["@activestate.com/koFileService;1"].
                               getService(Components.interfaces.koIFileService);
            return this.getServiceForFile(fileService.getFileFromURI(url));
        } catch (e) {
            log.exception(e);
        }
        return null;
    };


        /**
         * Internal diff helper. Used by both Diff and DiffRevisions.
         * @private
         *
         * @param commandType {string}  The scc command to be performed.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param additionalOptions {string}  Additional diff command options.
         * @param postCallback {function}  Function to be called on completion.
         * @param function_args {array}  The list of arguments for the command.
         */
    function _DiffHandler(commandType, sccSvc, additionalOptions,
                                     postCallback, function_args) {
        // We set ko_api namespace to reference the main Window as when
        // this is called from a dialog, and the user has their
        // diff preferences set to 'tab', then the opening of the new
        // diff tab will fail because the dialog will not have a view
        // manager or be able to show tabs, it needs to work through
        // the main Komodo view manager.
        var ko_api = ko.windowManager.getMainWindow().ko;
        var dialog;
        var cwd = null;
        var options = _getDiffOptions(sccSvc);
        var external = sccSvc.getValue("external_diff", "");
        var diffDisplayStyle = _prefs.getStringPref('diffDisplayStyle');

        if (additionalOptions) {
            options += " " + additionalOptions;
        }

        var async_callback = {
            "callback": function(result, data) {
                if (external) {
                    // Thanks for the info, but we have nothing to do...
                    return;
                }

                // Grab the variables we need from the settings.
                try {
                    if (!data) {
                        if (diffDisplayStyle == 'diffwin') {
                            dialog.close();
                        }
                        ko_api.dialogs.alert("Files are in sync with repository.",
                                             null, "Diff");
                        return;
                    }

                    if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                        if (diffDisplayStyle == 'diffwin') {
                            dialog.arguments[0].diff = data;
                            dialog.arguments[0].cwd = cwd;
                            dialog.arguments[0].async_op = null;
                            if (dialog._diffWindow) {
                                dialog.loadDiffResult(data, cwd);
                            } // else the dialog is not yet fully loaded, the
                              // load routines will call loadDiffResult later.

                        } else if (diffDisplayStyle == 'tab') {
                            ko_api.views.manager.doNewViewAsync('Diff', 'editor', function(view) {
                                view.koDoc.buffer = data;
                                view.koDoc.isDirty = false;
                            });
                        }

                    } else if (result == Components.interfaces.koIAsyncCallback.RESULT_STOPPED) {
                        if (diffDisplayStyle == 'diffwin') {
                            dialog.close();
                        }

                    } else if (result == Components.interfaces.koIAsyncCallback.RESULT_ERROR) {
                        // Only show errors for the internal diff tool.
                        if (diffDisplayStyle == 'diffwin') {
                            dialog.close();
                        }
                        // data should contain the error message then.
                        if (data) {
                            ko_api.dialogs.alert(data);
                        } else {
                            ko_api.dialogs.alert("Diff could not be retrieved.");
                        }
                    }
                } catch (e) {
                    log.exception(e, "could not load diff results");
                } finally {
                    // If supplied, trigger the callback function.
                    if (postCallback) {
                        postCallback(result, data);
                    }
                }
            }
        };

        var async_op;
        var title;

        if (commandType == 'Diff') {
            if (function_args.length == 1) {
                cwd = ko.uriparse.dirName(function_args);
            } else {
                cwd = ko.uriparse.commonURIPrefixFromURIs(function_args);
            }
            async_op = sccSvc.diff(function_args.length, function_args,
                                   options, external, async_callback);
            var urlBasenames = function_args.map(ko.uriparse.baseName);
            title = sccSvc.name + " diff: " + urlBasenames.join(", ")
        } else if (commandType == 'DiffRelative') {
            var baseURI = function_args[0];
            var relpaths = function_args[1];
            cwd = baseURI;
            async_op = sccSvc.diffRelative(baseURI,
                                           relpaths.length, relpaths,
                                           options, external, async_callback);
            var urlBasenames = function_args.map(ko.uriparse.baseName);
            title = sccSvc.name + " diff: " + urlBasenames.join(", ")
        } else if (commandType == 'DiffRevisions') {
            async_op = sccSvc.diffRevisions(function_args[0], function_args[1],
                                            function_args[2], function_args[3],
                                            function_args[4],
                                            options, external, async_callback);
            title = sccSvc.name + " diff: " +
                        ko.uriparse.baseName(function_args[0]) +
                        "#" + function_args[1] + ", " +
                        ko.uriparse.baseName(function_args[2]) +
                        "#" + function_args[3];
        }

        if (diffDisplayStyle == 'diffwin' && !external) {
            var obj = {
                "title": title,
                "diff": '',
                "async_op": async_op
            };
            dialog = ko.windowManager.openDialog(
                        "chrome://komodo/content/dialogs/diff.xul",
                        "_blank",
                        "chrome,all,close=yes,resizable=yes,scrollbars=yes",
                        obj);
        }
    };

        /**
         * Perform a diff revisions, and display the results.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls used.
         * @param additionalOptions {string}  Additional diff command options.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.Diff = function KoSccDiff(sccSvc, urls, additionalOptions, postCallback) {
        // XXX - if no sccSvc, work out what service it needs.
        try {
            _DiffHandler('Diff', sccSvc, additionalOptions, postCallback, urls);
        } catch (e) {
            log.exception(e);
        }
    }

        /**
         * Perform a diff revisions on the relative paths, display the results.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param baseURI {string}  The base uri to work from.
         * @param relpaths {array}  The list relative paths to diff.
         * @param additionalOptions {string}  Additional diff command options.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.DiffRelative = function KoSccDiff(sccSvc, baseURI, relpaths,
                                           additionalOptions, postCallback) {
        // XXX - if no sccSvc, work out what service it needs.
        try {
            _DiffHandler('DiffRelative', sccSvc, additionalOptions,
                         postCallback, [baseURI, relpaths]);
        } catch (e) {
            log.exception(e);
        }
    }

        /**
         * Perform a diff revisions, and display the results.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param uri1 {string}      The first file uri to diff with.
         * @param version1 {string}  Version of the first file to use.
         * @param uri1 {string}      The second file uri to diff against.
         * @param version2 {string}  Version of the second file to use.
         * @param filepath {string}  Local path of the file used to diff.
         * @param additionalOptions {string}  Additional diff command options.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.DiffRevisions = function KoSccDiffRevisions(sccSvc,
                                                     uri1, version1,
                                                     uri2, version2,
                                                     filepath,
                                                     additionalOptions,
                                                     postCallback) {
        // XXX - if no sccSvc, work out what service it needs.
        try {
            var function_args = [arguments[1], arguments[2], arguments[3],
                                 arguments[4], arguments[5]];
            _DiffHandler('DiffRevisions', sccSvc, additionalOptions,
                         postCallback, function_args);
        } catch (e) {
            log.exception(e);
        }
    }

        /**
         * Edit the supplied urls under SCC.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls to be edited.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.Edit = function(sccSvc, urls, postCallback) {
        _genericAsyncSccCommand(sccSvc, "edit", urls, postCallback);
    }

        /**
         * Add the supplied urls to SCC.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls to be added.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.Add = function(sccSvc, urls, postCallback) {
        _genericAsyncSccCommand(sccSvc, "add", urls, postCallback);
    }

        /**
         * Remove the supplied urls from SCC.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls to be removed.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.Remove = function(sccSvc, urls, postCallback) {
        var prompt = null;
        // XXX - This should come from the scc service.
        if (sccSvc.name == 'p4') {
            prompt = "Select the file you would like to remove from the " +
                     "repository. (The selected files will be removed from " +
                     "your disk.)";
        } else {
            prompt = "Select the file you would like to remove from the " +
                     "repository.";
        }
        var selectedURLs = ko.dialogs.selectFromList(
                "Remove Files from Source Control", // title
                prompt,
                urls,
                null, // selectionCondition (null means default)
                ko.uriparse.displayPath); // stringifier

        if (selectedURLs != null) { // null if the user cancelled the dialog
            _genericAsyncSccCommand(sccSvc, "remove", selectedURLs, postCallback);
        }
    }

        /**
         * Revert (undo changes for) the supplied urls in SCC.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls to be added.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.Revert = function(sccSvc, urls, postCallback) {
        log.debug("doRevert...");
        var selectedURLs = ko.dialogs.selectFromList(
                "Revert Files to Repository", // title
                "Select the files to revert to the repository version. "+
                    "(Changes in opened documents will be lost).",
                urls,
                null, // selectionCondition (null means default)
                ko.uriparse.displayPath); // stringifier

        if (selectedURLs != null) { // null if the user cancelled the dialog
            _genericAsyncSccCommand(sccSvc, "revert", selectedURLs, postCallback);
        }
    }

        /**
         * Commit the changes for the supplied urls into SCC.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls to be commited.
         * @param message {string}  The checkin message.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.Commit= function(sccSvc, urls, message, postCallback) {
        if (typeof(urls) == 'undefined') urls = null;
        if (typeof(message) == 'undefined') message = null;

        var commit_log = log;
        var onCloseCallback = function(commit_args) {
            var selected_urls = commit_args.urls;
            if (commit_args.retval != 'OK' || !selected_urls) {
                // Nothing to do.
                return;
            }
            var async_callback = {
                "callback": function(result, data) {
                    try {
                        // data will be an array of koISCCHistoryItem,
                        // throw it into the tree view.
                        if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                            //var notifybox = document.getElementById("scc_history_notificationbox");
                            //notifybox.removeAllNotifications(false /* false means to slide away */);
                            ko.scc.logger.reportSCCAction(sccSvc, "commit", selected_urls, data);
                            // Revert the urls as a commit message can change the file
                            // contents, such as when using svn properties.
                            ko.views.manager.revertViewsByURL(selected_urls);
                        } else if (result == Components.interfaces.koIAsyncCallback.RESULT_ERROR) {
                            // data should contain the error message then.
                            var title = sccSvc.name + " commit failure";
                            if (!data) {
                                data = "Commit could not be completed.";
                            }
    
                            if (ko.dialogs.customButtons("Re-open the commit dialog?",
                                                         ["Cancel", "Re-Open"],
                                                         "Re-Open",
                                                         data, title)
                                            == "Re-Open") {
                                ko.scc.Commit(sccSvc, selected_urls, commit_args.message);
                            }
                        }
                    } catch (e) {
                        commit_log.warn("commit callback failed, exception: " + e);
                    } finally {
                        commit_log = null;
                        // If supplied, trigger the callback function.
                        if (postCallback) {
                            postCallback(result, data);
                        }
                    }
                }
            }

            // Perform the asynchronous commit operation.
            sccSvc.commit(selected_urls.length, selected_urls,
                          commit_args.message, '', async_callback);
        }

        var obj = new Object();
        obj.urls = urls;
        obj.message = message;
        obj.sccSvc = sccSvc;
        obj.type = "commit";
        obj.callbackFn = onCloseCallback;

        // Show the dialog.
        ko.windowManager.openDialog("chrome://komodo/content/scc/commit.xul",
                                    "_blank",
                                    "chrome,titlebar,resizable=yes",
                                    obj);
    }

        /**
         * Load the SCC history dialog for the given urls.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls.
         * @param localRevisions {array}  The matching list of current scc file
         *                                versions, must be one for every url.
         */
    this.History = function(sccSvc, urls, localRevisions) {
        // Show the dialog.
        var obj = new Object();
        obj.urls = urls;
        obj.localRevisions = localRevisions;
        obj.sccHandler = sccSvc;
        ko.windowManager.openDialog(
            "chrome://komodo/content/scc/history.xul",
            "_blank",
            "chrome,titlebar,resizable=yes",
            obj);
    }

        /**
         * Update the supplied urls for latest SCC contents.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param urls {array}  The list of urls.
         * @param postCallback {function}  Function to be called on completion.
         */
    this.Update = function(sccSvc, urls, postCallback) {
        var _log = log;
        var async_callback = {
            "callback": function(result, data) {
                try {
                    // data will be an array of koISCCHistoryItem,
                    // throw it into the tree view.
                    if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                        ko.scc.logger.reportSCCAction(sccSvc, "update", urls, data);
                        ko.views.manager.revertViewsByURL(urls);
                        /* Force a refresh here: bug 85294, to refresh all
                           url's. XXX: This could be optimized later... */
                        ko.window.checkDiskFiles();
                        if (data) {
                            _handleChanges(data);
                        }
                    } else if (result == Components.interfaces.koIAsyncCallback.RESULT_ERROR) {
                        // data should contain the error message then.
                        var title = sccSvc.name + " update failure";
                        if (!data) {
                            data = "Update could not be completed.";
                        }
                        ko.dialogs.alert(title, data, title);
                    }
                } catch (e) {
                    _log.warn("update callback failed, exception: " + e);
                } finally {
                    _log = null;
                    // If supplied, trigger the callback function.
                    if (postCallback) {
                        postCallback(result, data);
                    }
                }
            }
        }

        sccSvc.update(urls.length, urls, '', async_callback);
    }

        /**
         * Checkout a SCC repository.
         * @param checkoutFolder {string} (optional) Initial checkout directory.
         */
    this.Checkout = function(checkoutFolder) {
        // Show the dialog.
        var obj = new Object();
        obj.checkoutFolder = checkoutFolder;
        ko.windowManager.openDialog(
            "chrome://komodo/content/scc/checkout.xul",
            "_blank",
            "chrome,titlebar,resizable=yes",
            obj);
    }

        /**
         * Warn the user about enabling CVS and SVN SSH integration.
         */
    this.warnSSHConfiguration = function SCCWarnAboutSSHIfNecessary() {
        var osSvc = Components.classes["@activestate.com/koOs;1"].getService();
        var CVS_RSH = osSvc.getenv('CVS_RSH');
        var SVN_SSH = osSvc.getenv('SVN_SSH');
        var msg, answer;
        // Only bother warning if CVS_RSH is set and CVS integration and CVS
        // background checking are enabled.
        var _prefs = Components.classes["@activestate.com/koPrefService;1"].
                        getService(Components.interfaces.koIPrefService).prefs;
        if (CVS_RSH
            && _prefs.getBooleanPref("cvsEnabled")
            && _prefs.getBooleanPref("cvsBackgroundCheck")
            // Checking this pref (seemingly redundantly) here is necessary
            // because the pref is being used to mean:
            //  "don't ask or do anything if true"
            // rather than the usual meaning:
            //  "don't ask, just presume user wants same action as last time"
            && !_prefs.getBooleanPref("donotask_cvs_ssh_setup_warning")
           )
        {
            msg = "The CVS_RSH environment variable is set to '"+CVS_RSH+
                "' and CVS Background Status Checking is currently turned "+
                "on.  Failure to properly configure CVS authentication may "+
                "result in process lockups within Komodo.  Please read the "+
                "Komodo documentation regarding the use of external "+
                "protocols, such as SSH, with Komodo.  \nIf you are sure "+
                "you have properly configured CVS/SSH authentication on "+
                "your machine, click 'Keep Enabled'.  Otherwise click "+
                "'Disable' to disable Komodo's CVS integration for now. "+
                "When you are sure you have setup CVS properly you can "+
                "re-enable CVS integration in Edit|Preferences|Source Control|CVS.";
            answer = ko.dialogs.customButtons(msg,
                ["Disable", "Disable and Tell Me More", "Keep Enabled"],
                "Disable", null, "Komodo CVS Integration",
                "cvs_ssh_setup_warning");
            switch (answer) {
                case "Keep Enabled":
                    break;
                case "Disable and Tell Me More":
                    ko.help.open("scc.html#config_cvs");
                    // fall through
                case "Disable":
                case "Cancel":
                    _prefs.setBooleanPref("cvsEnabled", false);
            }
        }

        if (SVN_SSH
            && _prefs.getBooleanPref("svnEnabled")
            && _prefs.getBooleanPref("svnBackgroundCheck")
            // Checking this pref (seemingly redundantly) here is necessary
            // because the pref is being used to mean:
            //  "don't ask or do anything if true"
            // rather than the usual meaning:
            //  "don't ask, just presume user wants same action as last time"
            && !_prefs.getBooleanPref("donotask_svn_ssh_setup_warning")
           )
        {
            msg = "The SVN_SSH environment variable is set to '"+SVN_SSH+
                      "' and Subversion Background Status Checking is "+
                      "currently turned on.  Failure to properly "+
                      "configure Subversion authentication may result "+
                      "in process lockups within Komodo.  Please read "+
                      "the Komodo documentation regarding the use of "+
                      "external protocols, such as SSH, with Komodo.  \n"+
                      "If you are sure you have properly configured "+
                      "Subversion/SSH authentication on your machine, "+
                      "click 'Keep Enabled'.  Otherwise click 'Disable' "+
                      "to disable Komodo's Subversion integration for "+
                      "now. When you are sure you have setup Subversion "+
                      "properly you can re-enable Subversion integration "+
                      "in Edit|Preferences|Source Control|Subversion.";
            answer = ko.dialogs.customButtons(msg,
                ["Disable", "Disable and Tell Me More", "Keep Enabled"],
                "Disable", null, "Komodo Subversion Integration",
                "svn_ssh_setup_warning");
            switch (answer) {
                case "Keep Enabled":
                    break;
                case "Disable and Tell Me More":
                    ko.help.open("scc.html#config_svn");
                    // fall through
                case "Disable":
                case "Cancel":
                    _prefs.setBooleanPref("svnEnabled", false);
            }
        }
    }
}).apply(ko.scc);

/* For the main Komodo SCC scintilla widget - logging handling. */
ko.scc.logger = {};

(function() {

        /**
         * One off initialization of the SCC output panel text.
         */
    this.init = function SCCScinInit() {
        try {
            var scin = document.getElementById('scin-sccoutput');
            // The scc output window will not be available from a Komodo dialog.
            if (scin) {
                scin.initWithBuffer('', 'Text');
            }
        } finally {
            // Now that we are initialized, add a finalizer as well.
            window.addEventListener("unload", ko.scc.logger.finalize, false);
        }
    }

    this.finalize = function SCCScinFinalize() {
        var scin = document.getElementById('scin-sccoutput');
        // The scc output window will not be available from a Komodo dialog.
        if (scin) {
            scin.close();
        }
    }

        /**
         * Add the description and text to the SCC output panel.
         * @param actiondescription {string}  Title description of text.
         * @param actiontext {string}  The text to add to the SCC output panel.
         */
    this.reportResults = function SCCScinReportResults(actiondescription, actiontext)
    {
        var mainWindow = ko.windowManager.getMainWindow();
        mainWindow.document.getElementById('sccoutput-desc').setAttribute('value', actiondescription);
        this.show();
        this.append(actiontext);
    }

        /**
         * Report the SCC command results in the main Komodo SCC output panel.
         * @param sccSvc {Components.interfaces.koISCC}  SCC service instance.
         * @param cmdName {string}  The name of the scc command.
         * @param urls {array}  The list of urls used.
         * @param text {string}  The command output text.
         */
    this.reportSCCAction = function SCCScinReportAction(sccSvc, cmdName, urls, text)
    {
        var title = "Last command: " + cmdName + ' ('+sccSvc.name + ') on ';
        var display = ko.uriparse.displayPath(urls[0]);
        if (urls.length > 1) {
            title += ' ' + display + ', ... (' + urls.length + ' files/directories).';
        } else {
            title += ' ' + display + '.';
        }

        var executable = sccSvc.executable;
        var output = (executable ? executable : sccSvc.name) + " " + cmdName+'\n';
        for (var url in urls) {
            output += '    '+urls[url]+'\n'
        }
        output += text;

        this.reportResults(title, output);
    }

        /**
         * Show the main Komodo SCC output panel, will ask if the user wants
         * to see it.
         */
    this.show = function SCCTab_Show()
    {
        var mainWindow = ko.windowManager.getMainWindow();
        var tabWidget = mainWindow.document.getElementById("sccoutput_tab");
        var tabsWidget = mainWindow.document.getElementById("output_tabs");
        if (! ko.uilayout.outputPaneShown() || tabsWidget.selectedItem != tabWidget) {
            if (ko.dialogs.yesNo("Show SCC Output Tab?", // prompt
                             "Yes", // response
                             null, // text
                             null, // title
                             'show_scc_tab') == "Yes") {
                ko.uilayout.ensureOutputPaneShown();
                tabsWidget.selectedItem = tabWidget;
            }
        }
    }

        /**
         * Adds text to the SCC output panel.
         * @param text {string}  The output text to be added.
         */
    this.append = function SCCScinAppendText(text) {
        if (typeof(text)=='undefined' || !text) return;
        var mainWindow = ko.windowManager.getMainWindow();
        var scin = mainWindow.document.getElementById('scin-sccoutput').scimoz;
        scin.readOnly = false;
        scin.currentPos = scin.textLength-1;
        // move cursor to end of current text
        text = "-------------------------------------------------------------------\n" +
            text+'\n';
        scin.addText(text.length, text);
        scin.gotoLine(scin.lineFromPosition(scin.textLength));
        scin.readOnly = true;
    }

        /**
         * Clear all the text in the SCC output panel.
         */
    this.clear = function SCCScinClearText() {
        var mainWindow = ko.windowManager.getMainWindow();
        var scin = mainWindow.document.getElementById('scin-sccoutput').scimoz;
        scin.readOnly = false;
        scin.text = '';
        scin.readOnly = true;
    }

}).apply(ko.scc.logger);

window.addEventListener("load", ko.scc.logger.init, false);
